import click

import idom
from idom._console.rewrite_camel_case_props import rewrite_camel_case_props
from idom._console.rewrite_keys import rewrite_keys


@click.group()
@click.version_option(idom.__version__, prog_name=idom.__name__)
def app() -> None:
    pass


app.add_command(rewrite_keys)
app.add_command(rewrite_camel_case_props)


if __name__ == "__main__":
    app()



from __future__ import annotations

import ast
from collections.abc import Sequence
from dataclasses import dataclass
from pathlib import Path
from textwrap import indent
from tokenize import COMMENT as COMMENT_TOKEN
from tokenize import generate_tokens
from typing import Any, Iterator

import click

from idom import html


def rewrite_changed_nodes(
    file: Path,
    source: str,
    tree: ast.AST,
    changed: list[ChangedNode],
) -> str:
    ast.fix_missing_locations(tree)

    lines = source.split("\n")

    # find closest parent nodes that should be re-written
    nodes_to_unparse: list[ast.AST] = []
    for change in changed:
        node_lineage = [change.node, *change.parents]
        for i in range(len(node_lineage) - 1):
            current_node, next_node = node_lineage[i : i + 2]
            if (
                not hasattr(next_node, "lineno")
                or next_node.lineno < change.node.lineno
                or isinstance(next_node, (ast.ClassDef, ast.FunctionDef))
            ):
                nodes_to_unparse.append(current_node)
                break
        else:  # pragma: no cover
            raise RuntimeError("Failed to change code")

    # check if an nodes to rewrite contain eachother, pick outermost nodes
    current_outermost_node, *sorted_nodes_to_unparse = list(
        sorted(nodes_to_unparse, key=lambda n: n.lineno)
    )
    outermost_nodes_to_unparse = [current_outermost_node]
    for node in sorted_nodes_to_unparse:
        if (
            not current_outermost_node.end_lineno
            or node.lineno > current_outermost_node.end_lineno
        ):
            current_outermost_node = node
            outermost_nodes_to_unparse.append(node)

    moved_comment_lines_from_end: list[int] = []
    # now actually rewrite these nodes (in reverse to avoid changes earlier in file)
    for node in reversed(outermost_nodes_to_unparse):
        # make a best effort to preserve any comments that we're going to overwrite
        comments = _find_comments(lines[node.lineno - 1 : node.end_lineno])

        # there may be some content just before and after the content we're re-writing
        before_replacement = lines[node.lineno - 1][: node.col_offset].lstrip()

        after_replacement = (
            lines[node.end_lineno - 1][node.end_col_offset :].strip()
            if node.end_lineno is not None and node.end_col_offset is not None
            else ""
        )

        replacement = indent(
            before_replacement
            + "\n".join([*comments, ast.unparse(node)])
            + after_replacement,
            " " * (node.col_offset - len(before_replacement)),
        )

        lines[node.lineno - 1 : node.end_lineno or node.lineno] = [replacement]

        if comments:
            moved_comment_lines_from_end.append(len(lines) - node.lineno)

    for lineno_from_end in sorted(list(set(moved_comment_lines_from_end))):
        click.echo(f"Moved comments to {file}:{len(lines) - lineno_from_end}")

    return "\n".join(lines)


@dataclass
class ChangedNode:
    node: ast.AST
    parents: Sequence[ast.AST]


def find_element_constructor_usages(
    tree: ast.AST, add_props: bool = False
) -> Iterator[ElementConstructorInfo]:
    changed: list[Sequence[ast.AST]] = []
    for parents, node in _walk_with_parent(tree):
        if not (isinstance(node, ast.Call)):
            continue

        func = node.func
        if (
            isinstance(func, ast.Attribute)
            and isinstance(func.value, ast.Name)
            and func.value.id == "html"
        ):
            name = func.attr
        elif isinstance(func, ast.Name):
            name = func.id
        else:
            continue

        maybe_attr_dict_node: Any | None = None

        if name == "vdom":
            if len(node.args) == 0:
                continue
            elif len(node.args) == 1:
                maybe_attr_dict_node = ast.Dict(keys=[], values=[])
                if add_props:
                    node.args.append(maybe_attr_dict_node)
                else:
                    continue
            elif isinstance(node.args[1], (ast.Constant, ast.JoinedStr)):
                maybe_attr_dict_node = ast.Dict(keys=[], values=[])
                if add_props:
                    node.args.insert(1, maybe_attr_dict_node)
                else:
                    continue
            elif len(node.args) >= 2:
                maybe_attr_dict_node = node.args[1]
        elif hasattr(html, name):
            if len(node.args) == 0:
                maybe_attr_dict_node = ast.Dict(keys=[], values=[])
                if add_props:
                    node.args.append(maybe_attr_dict_node)
                else:
                    continue
            elif isinstance(node.args[0], (ast.Constant, ast.JoinedStr)):
                maybe_attr_dict_node = ast.Dict(keys=[], values=[])
                if add_props:
                    node.args.insert(0, maybe_attr_dict_node)
                else:
                    continue
            else:
                maybe_attr_dict_node = node.args[0]

        if not maybe_attr_dict_node:
            continue

        if isinstance(maybe_attr_dict_node, ast.Dict) or (
            isinstance(maybe_attr_dict_node, ast.Call)
            and isinstance(maybe_attr_dict_node.func, ast.Name)
            and maybe_attr_dict_node.func.id == "dict"
            and isinstance(maybe_attr_dict_node.func.ctx, ast.Load)
        ):
            yield ElementConstructorInfo(node, maybe_attr_dict_node, parents)

    return changed


@dataclass
class ElementConstructorInfo:
    call: ast.Call
    props: ast.Dict | ast.Call
    parents: Sequence[ast.AST]


def _find_comments(lines: list[str]) -> list[str]:
    iter_lines = iter(lines)
    return [
        token
        for token_type, token, _, _, _ in generate_tokens(lambda: next(iter_lines))
        if token_type == COMMENT_TOKEN
    ]


def _walk_with_parent(
    node: ast.AST, parents: tuple[ast.AST, ...] = ()
) -> Iterator[tuple[tuple[ast.AST, ...], ast.AST]]:
    parents = (node,) + parents
    for child in ast.iter_child_nodes(node):
        yield parents, child
        yield from _walk_with_parent(child, parents)



from base64 import b64encode
from pathlib import Path

import idom
from idom.testing import DisplayFixture, poll
from tests.tooling.common import DEFAULT_TYPE_DELAY


HERE = Path(__file__).parent


IMAGE_SRC_BYTES = b"""
<svg width="400" height="110" xmlns="http://www.w3.org/2000/svg">
    <rect width="300" height="100" style="fill:rgb(0,0,255);" />
</svg>
"""
BASE64_IMAGE_SRC = b64encode(IMAGE_SRC_BYTES).decode()


async def test_image_from_string(display: DisplayFixture):
    src = IMAGE_SRC_BYTES.decode()
    await display.show(lambda: idom.widgets.image("svg", src, {"id": "a-circle-1"}))
    client_img = await display.page.wait_for_selector("#a-circle-1")
    assert BASE64_IMAGE_SRC in (await client_img.get_attribute("src"))


async def test_image_from_bytes(display: DisplayFixture):
    src = IMAGE_SRC_BYTES
    await display.show(lambda: idom.widgets.image("svg", src, {"id": "a-circle-1"}))
    client_img = await display.page.wait_for_selector("#a-circle-1")
    assert BASE64_IMAGE_SRC in (await client_img.get_attribute("src"))


async def test_use_linked_inputs(display: DisplayFixture):
    @idom.component
    def SomeComponent():
        i_1, i_2 = idom.widgets.use_linked_inputs([{"id": "i_1"}, {"id": "i_2"}])
        return idom.html.div(i_1, i_2)

    await display.show(SomeComponent)

    input_1 = await display.page.wait_for_selector("#i_1")
    input_2 = await display.page.wait_for_selector("#i_2")

    await input_1.type("hello", delay=DEFAULT_TYPE_DELAY)

    assert (await input_1.evaluate("e => e.value")) == "hello"
    assert (await input_2.evaluate("e => e.value")) == "hello"

    await input_2.focus()
    await input_2.type(" world", delay=DEFAULT_TYPE_DELAY)

    assert (await input_1.evaluate("e => e.value")) == "hello world"
    assert (await input_2.evaluate("e => e.value")) == "hello world"


async def test_use_linked_inputs_on_change(display: DisplayFixture):
    value = idom.Ref(None)

    @idom.component
    def SomeComponent():
        i_1, i_2 = idom.widgets.use_linked_inputs(
            [{"id": "i_1"}, {"id": "i_2"}],
            on_change=value.set_current,
        )
        return idom.html.div(i_1, i_2)

    await display.show(SomeComponent)

    input_1 = await display.page.wait_for_selector("#i_1")
    input_2 = await display.page.wait_for_selector("#i_2")

    await input_1.type("hello", delay=DEFAULT_TYPE_DELAY)

    poll_value = poll(lambda: value.current)

    await poll_value.until_equals("hello")

    await input_2.focus()
    await input_2.type(" world", delay=DEFAULT_TYPE_DELAY)

    await poll_value.until_equals("hello world")


async def test_use_linked_inputs_on_change_with_cast(display: DisplayFixture):
    value = idom.Ref(None)

    @idom.component
    def SomeComponent():
        i_1, i_2 = idom.widgets.use_linked_inputs(
            [{"id": "i_1"}, {"id": "i_2"}], on_change=value.set_current, cast=int
        )
        return idom.html.div(i_1, i_2)

    await display.show(SomeComponent)

    input_1 = await display.page.wait_for_selector("#i_1")
    input_2 = await display.page.wait_for_selector("#i_2")

    await input_1.type("1")

    poll_value = poll(lambda: value.current)

    await poll_value.until_equals(1)

    await input_2.focus()
    await input_2.type("2")

    await poll_value.until_equals(12)


async def test_use_linked_inputs_ignore_empty(display: DisplayFixture):
    value = idom.Ref(None)

    @idom.component
    def SomeComponent():
        i_1, i_2 = idom.widgets.use_linked_inputs(
            [{"id": "i_1"}, {"id": "i_2"}],
            on_change=value.set_current,
            ignore_empty=True,
        )
        return idom.html.div(i_1, i_2)

    await display.show(SomeComponent)

    input_1 = await display.page.wait_for_selector("#i_1")
    input_2 = await display.page.wait_for_selector("#i_2")

    await input_1.type("1")

    poll_value = poll(lambda: value.current)

    await poll_value.until_equals("1")

    await input_2.focus()
    await input_2.press("Backspace")

    await poll_value.until_equals("1")

    await input_2.type("2")

    await poll_value.until_equals("2")



import os
from unittest import mock

import pytest

from idom._option import DeprecatedOption, Option


def test_option_repr():
    opt = Option("A_FAKE_OPTION", "some-value")
    assert opt.name == "A_FAKE_OPTION"
    assert repr(opt) == "Option(A_FAKE_OPTION='some-value')"


@mock.patch.dict(os.environ, {"A_FAKE_OPTION": "value-from-environ"})
def test_option_from_os_environ():
    opt = Option("A_FAKE_OPTION", "default-value")
    assert opt.current == "value-from-environ"


def test_option_from_default():
    opt = Option("A_FAKE_OPTION", "default-value")
    assert opt.current == "default-value"
    assert opt.current is opt.default


@mock.patch.dict(os.environ, {"A_FAKE_OPTION": "1"})
def test_option_validator():
    opt = Option("A_FAKE_OPTION", False, validator=lambda x: bool(int(x)))

    assert opt.current is True

    opt.current = "0"
    assert opt.current is False

    with pytest.raises(ValueError, match="invalid literal for int"):
        opt.current = "not-an-int"


def test_immutable_option():
    opt = Option("A_FAKE_OPTION", "default-value", mutable=False)
    assert not opt.mutable
    with pytest.raises(TypeError, match="cannot be modified after initial load"):
        opt.current = "a-new-value"
    with pytest.raises(TypeError, match="cannot be modified after initial load"):
        opt.unset()


def test_option_reset():
    opt = Option("A_FAKE_OPTION", "default-value")
    opt.current = "a-new-value"
    opt.unset()
    assert opt.current is opt.default
    assert not opt.is_set()


@mock.patch.dict(os.environ, {"A_FAKE_OPTION": "value-from-environ"})
def test_option_reload():
    opt = Option("A_FAKE_OPTION", "default-value")
    opt.current = "some-other-value"
    opt.reload()
    assert opt.current == "value-from-environ"


def test_option_set():
    opt = Option("A_FAKE_OPTION", "default-value")
    assert not opt.is_set()
    opt.current = "a-new-value"
    assert opt.is_set()


def test_option_set_default():
    opt = Option("A_FAKE_OPTION", "default-value")
    assert not opt.is_set()
    assert opt.set_default("new-value") == "new-value"
    assert opt.is_set()


def test_cannot_subscribe_immutable_option():
    opt = Option("A_FAKE_OPTION", "default", mutable=False)
    with pytest.raises(TypeError, match="Immutable options cannot be subscribed to"):
        opt.subscribe(lambda value: None)


def test_option_subscribe():
    opt = Option("A_FAKE_OPTION", "default")

    calls = []
    opt.subscribe(calls.append)
    assert calls == ["default"]

    opt.current = "default"
    # value did not change, so no trigger
    assert calls == ["default"]

    opt.current = "new-1"
    opt.current = "new-2"
    assert calls == ["default", "new-1", "new-2"]

    opt.unset()
    assert calls == ["default", "new-1", "new-2", "default"]


def test_deprecated_option():
    opt = DeprecatedOption("is deprecated!", "A_FAKE_OPTION", None)

    with pytest.warns(DeprecationWarning, match="is deprecated!"):
        opt.current

    with pytest.warns(DeprecationWarning, match="is deprecated!"):
        opt.current = "something"



from idom.sample import SampleApp
from idom.testing import DisplayFixture


async def test_sample_app(display: DisplayFixture):
    await display.show(SampleApp)

    h1 = await display.page.wait_for_selector("h1")
    assert (await h1.text_content()) == "Sample Application"



import pytest

from idom import config
from idom._option import Option


@pytest.fixture(autouse=True)
def reset_options():
    options = [value for value in config.__dict__.values() if isinstance(value, Option)]

    should_unset = object()
    original_values = []
    for opt in options:
        original_values.append(opt.current if opt.is_set() else should_unset)

    yield

    for opt, val in zip(options, original_values):
        if val is should_unset:
            if opt.is_set():
                opt.unset()
        else:
            opt.current = val


def test_idom_debug_mode_toggle():
    # just check that nothing breaks
    config.IDOM_DEBUG_MODE.current = True
    config.IDOM_DEBUG_MODE.current = False



from html import escape as html_escape

import pytest

import idom
from idom import html
from idom.utils import (
    HTMLParseError,
    del_html_head_body_transform,
    html_to_vdom,
    vdom_to_html,
)


def test_basic_ref_behavior():
    r = idom.Ref(1)
    assert r.current == 1

    r.current = 2
    assert r.current == 2

    assert r.set_current(3) == 2
    assert r.current == 3

    r = idom.Ref()
    with pytest.raises(AttributeError):
        r.current

    r.current = 4
    assert r.current == 4


def test_ref_equivalence():
    assert idom.Ref([1, 2, 3]) == idom.Ref([1, 2, 3])
    assert idom.Ref([1, 2, 3]) != idom.Ref([1, 2])
    assert idom.Ref([1, 2, 3]) != [1, 2, 3]
    assert idom.Ref() != idom.Ref()
    assert idom.Ref() != idom.Ref(1)


def test_ref_repr():
    assert repr(idom.Ref([1, 2, 3])) == "Ref([1, 2, 3])"
    assert repr(idom.Ref()) == "Ref(<undefined>)"


@pytest.mark.parametrize(
    "case",
    [
        {"source": "<div/>", "model": {"tagName": "div"}},
        {
            "source": "<div style='background-color:blue'/>",
            "model": {
                "tagName": "div",
                "attributes": {"style": {"background_color": "blue"}},
            },
        },
        {
            "source": "<div>Hello!</div>",
            "model": {"tagName": "div", "children": ["Hello!"]},
        },
        {
            "source": "<div>Hello!<p>World!</p></div>",
            "model": {
                "tagName": "div",
                "children": ["Hello!", {"tagName": "p", "children": ["World!"]}],
            },
        },
    ],
)
def test_html_to_vdom(case):
    assert html_to_vdom(case["source"]) == case["model"]


def test_html_to_vdom_transform():
    source = "<p>hello <a>world</a> and <a>universe</a>lmao</p>"

    def make_links_blue(node):
        if node["tagName"] == "a":
            node["attributes"] = {"style": {"color": "blue"}}
        return node

    expected = {
        "tagName": "p",
        "children": [
            "hello ",
            {
                "tagName": "a",
                "children": ["world"],
                "attributes": {"style": {"color": "blue"}},
            },
            " and ",
            {
                "tagName": "a",
                "children": ["universe"],
                "attributes": {"style": {"color": "blue"}},
            },
            "lmao",
        ],
    }

    assert html_to_vdom(source, make_links_blue) == expected


def test_non_html_tag_behavior():
    source = "<my-tag data-x=something><my-other-tag key=a-key /></my-tag>"

    expected = {
        "tagName": "my-tag",
        "attributes": {"data-x": "something"},
        "children": [
            {"tagName": "my-other-tag", "key": "a-key"},
        ],
    }

    assert html_to_vdom(source, strict=False) == expected

    with pytest.raises(HTMLParseError):
        html_to_vdom(source, strict=True)


def test_html_to_vdom_with_null_tag():
    source = "<p>hello<br>world</p>"

    expected = {
        "tagName": "p",
        "children": [
            "hello",
            {"tagName": "br"},
            "world",
        ],
    }

    assert html_to_vdom(source) == expected


def test_html_to_vdom_with_style_attr():
    source = '<p style="color: red; background-color : green; ">Hello World.</p>'

    expected = {
        "attributes": {"style": {"background_color": "green", "color": "red"}},
        "children": ["Hello World."],
        "tagName": "p",
    }

    assert html_to_vdom(source) == expected


def test_html_to_vdom_with_no_parent_node():
    source = "<p>Hello</p><div>World</div>"

    expected = {
        "tagName": "div",
        "children": [
            {"tagName": "p", "children": ["Hello"]},
            {"tagName": "div", "children": ["World"]},
        ],
    }

    assert html_to_vdom(source) == expected


def test_del_html_body_transform():
    source = """
    <!DOCTYPE html>
    <html lang="en">

    <head>
    <title>My Title</title>
    </head>

    <body><h1>Hello World</h1></body>

    </html>
    """

    expected = {
        "tagName": "",
        "children": [
            {
                "tagName": "",
                "children": [{"tagName": "title", "children": ["My Title"]}],
            },
            {
                "tagName": "",
                "children": [{"tagName": "h1", "children": ["Hello World"]}],
            },
        ],
    }

    assert html_to_vdom(source, del_html_head_body_transform) == expected


SOME_OBJECT = object()


@pytest.mark.parametrize(
    "vdom_in, html_out",
    [
        (
            html.div("hello"),
            "<div>hello</div>",
        ),
        (
            html.div(SOME_OBJECT),
            f"<div>{html_escape(str(SOME_OBJECT))}</div>",
        ),
        (
            html.div({"someAttribute": SOME_OBJECT}),
            f'<div someattribute="{html_escape(str(SOME_OBJECT))}"></div>',
        ),
        (
            html.div(
                "hello", html.a({"href": "https://example.com"}, "example"), "world"
            ),
            '<div>hello<a href="https://example.com">example</a>world</div>',
        ),
        (
            html.button({"on_click": lambda event: None}),
            "<button></button>",
        ),
        (
            html._("hello ", html._("world")),
            "hello world",
        ),
        (
            html._(html.div("hello"), html._("world")),
            "<div>hello</div>world",
        ),
        (
            html.div({"style": {"backgroundColor": "blue", "marginLeft": "10px"}}),
            '<div style="background-color:blue;margin-left:10px"></div>',
        ),
        (
            html.div({"style": "background-color:blue;margin-left:10px"}),
            '<div style="background-color:blue;margin-left:10px"></div>',
        ),
        (
            html._(
                html.div("hello"),
                html.a({"href": "https://example.com"}, "example"),
            ),
            '<div>hello</div><a href="https://example.com">example</a>',
        ),
        (
            html.div(
                html._(
                    html.div("hello"),
                    html.a({"href": "https://example.com"}, "example"),
                ),
                html.button(),
            ),
            '<div><div>hello</div><a href="https://example.com">example</a><button></button></div>',
        ),
        (
            html.div(
                {"data_something": 1, "data_something_else": 2, "dataisnotdashed": 3}
            ),
            '<div data-something="1" data-something-else="2" dataisnotdashed="3"></div>',
        ),
        (
            html.div(
                {"dataSomething": 1, "dataSomethingElse": 2, "dataisnotdashed": 3}
            ),
            '<div data-something="1" data-something-else="2" dataisnotdashed="3"></div>',
        ),
    ],
)
def test_vdom_to_html(vdom_in, html_out):
    assert vdom_to_html(vdom_in) == html_out


def test_vdom_to_html_error():
    with pytest.raises(TypeError, match="Expected a VDOM dict"):
        vdom_to_html({"notVdom": True})



from __future__ import annotations

import ast
import re
import sys
from keyword import kwlist
from pathlib import Path

import click

from idom._console.ast_utils import (
    ChangedNode,
    find_element_constructor_usages,
    rewrite_changed_nodes,
)


CAMEL_CASE_SUB_PATTERN = re.compile(r"(?<!^)(?=[A-Z])")


@click.command()
@click.argument("paths", nargs=-1, type=click.Path(exists=True))
def rewrite_camel_case_props(paths: list[str]) -> None:
    """Rewrite camelCase props to snake_case"""
    if sys.version_info < (3, 9):  # pragma: no cover
        raise RuntimeError("This command requires Python>=3.9")

    for p in map(Path, paths):
        for f in [p] if p.is_file() else p.rglob("*.py"):
            result = generate_rewrite(file=f, source=f.read_text())
            if result is not None:
                f.write_text(result)


def generate_rewrite(file: Path, source: str) -> str | None:
    tree = ast.parse(source)

    changed = find_nodes_to_change(tree)
    if not changed:
        return None

    new = rewrite_changed_nodes(file, source, tree, changed)
    return new


def find_nodes_to_change(tree: ast.AST) -> list[ChangedNode]:
    changed: list[ChangedNode] = []
    for el_info in find_element_constructor_usages(tree):
        if isinstance(el_info.props, ast.Dict):
            did_change = False
            keys: list[ast.expr | None] = []
            for k in el_info.props.keys:
                if isinstance(k, ast.Constant) and isinstance(k.value, str):
                    new_prop_name = conv_attr_name(k.value)
                    if new_prop_name != k.value:
                        did_change = True
                        keys.append(ast.Constant(new_prop_name))
                    else:
                        keys.append(k)
                else:
                    keys.append(k)
            if not did_change:
                continue
            el_info.props.keys = keys
        else:
            did_change = False
            keywords: list[ast.keyword] = []
            for kw in el_info.props.keywords:
                if kw.arg is not None:
                    new_prop_name = conv_attr_name(kw.arg)
                    if new_prop_name != kw.arg:
                        did_change = True
                        keywords.append(ast.keyword(arg=new_prop_name, value=kw.value))
                    else:
                        keywords.append(kw)
                else:
                    keywords.append(kw)
            if not did_change:
                continue
            el_info.props.keywords = keywords

        changed.append(ChangedNode(el_info.call, el_info.parents))
    return changed


def conv_attr_name(name: str) -> str:
    new_name = CAMEL_CASE_SUB_PATTERN.sub("_", name).lower()
    return f"{new_name}_" if new_name in kwlist else new_name



import asyncio
from contextlib import AsyncExitStack
from pathlib import Path

from playwright.async_api import Browser

import idom
from idom.backend.utils import find_available_port
from idom.testing import BackendFixture, DisplayFixture, poll
from tests.tooling.common import DEFAULT_TYPE_DELAY
from tests.tooling.hooks import use_counter


JS_DIR = Path(__file__).parent / "js"


async def test_automatic_reconnect(browser: Browser):
    port = find_available_port("localhost")
    page = await browser.new_page()

    # we need to wait longer here because the automatic reconnect is not instant
    page.set_default_timeout(10000)

    @idom.component
    def SomeComponent():
        count, incr_count = use_counter(0)
        return idom.html._(
            idom.html.p({"data_count": count, "id": "count"}, "count", count),
            idom.html.button(dict(on_click=lambda e: incr_count(), id="incr"), "incr"),
        )

    async with AsyncExitStack() as exit_stack:
        server = await exit_stack.enter_async_context(BackendFixture(port=port))
        display = await exit_stack.enter_async_context(
            DisplayFixture(server, driver=page)
        )

        await display.show(SomeComponent)

        count = await page.wait_for_selector("#count")
        incr = await page.wait_for_selector("#incr")

        for i in range(3):
            assert (await count.get_attribute("data-count")) == str(i)
            await incr.click()

    # the server is disconnected but the last view state is still shown
    await page.wait_for_selector("#count")

    async with AsyncExitStack() as exit_stack:
        server = await exit_stack.enter_async_context(BackendFixture(port=port))
        display = await exit_stack.enter_async_context(
            DisplayFixture(server, driver=page)
        )

        # use mount instead of show to avoid a page refesh
        display.backend.mount(SomeComponent)

        async def get_count():
            # need to refetch element because may unmount on reconnect
            count = await page.wait_for_selector("#count")
            return await count.get_attribute("data-count")

        for i in range(3):
            # it may take a moment for the websocket to reconnect so need to poll
            await poll(get_count).until_equals(str(i))

            # need to refetch element because may unmount on reconnect
            incr = await page.wait_for_selector("#incr")

            await incr.click()


async def test_style_can_be_changed(display: DisplayFixture):
    """This test was introduced to verify the client does not mutate the model

    A bug was introduced where the client-side model was mutated and React was relying
    on the model to have been copied in order to determine if something had changed.

    See for more info: https://github.com/idom-team/idom/issues/480
    """

    @idom.component
    def ButtonWithChangingColor():
        color_toggle, set_color_toggle = idom.hooks.use_state(True)
        color = "red" if color_toggle else "blue"
        return idom.html.button(
            {
                "id": "my-button",
                "onClick": lambda event: set_color_toggle(not color_toggle),
                "style": {"backgroundColor": color, "color": "white"},
            },
            f"color: {color}",
        )

    await display.show(ButtonWithChangingColor)

    button = await display.page.wait_for_selector("#my-button")

    assert (await _get_style(button))["background-color"] == "red"

    for color in ["blue", "red"] * 2:
        await button.click()
        assert (await _get_style(button))["background-color"] == color


async def _get_style(element):
    items = (await element.get_attribute("style")).split(";")
    pairs = [item.split(":", 1) for item in map(str.strip, items) if item]
    return {key.strip(): value.strip() for key, value in pairs}


async def test_slow_server_response_on_input_change(display: DisplayFixture):
    """A delay server-side could cause input values to be overwritten.

    For more info see: https://github.com/idom-team/idom/issues/684
    """

    delay = 0.2

    @idom.component
    def SomeComponent():
        value, set_value = idom.hooks.use_state("")

        async def handle_change(event):
            await asyncio.sleep(delay)
            set_value(event["target"]["value"])

        return idom.html.input({"onChange": handle_change, "id": "test-input"})

    await display.show(SomeComponent)

    inp = await display.page.wait_for_selector("#test-input")
    await inp.type("hello", delay=DEFAULT_TYPE_DELAY)

    assert (await inp.evaluate("node => node.value")) == "hello"


async def test_snake_case_attributes(display: DisplayFixture):
    @idom.component
    def SomeComponent():
        return idom.html.h1(
            {
                "id": "my-title",
                "style": {"background_color": "blue"},
                "class_name": "hello",
                "data_some_thing": "some-data",
                "aria_some_thing": "some-aria",
            },
            "title with some attributes",
        )

    await display.show(SomeComponent)

    title = await display.page.wait_for_selector("#my-title")

    assert await title.get_attribute("class") == "hello"
    assert await title.get_attribute("style") == "background-color: blue;"
    assert await title.get_attribute("data-some-thing") == "some-data"
    assert await title.get_attribute("aria-some-thing") == "some-aria"



import logging
import os

import pytest

from idom import Ref, component, html, testing
from idom.backend import starlette as starlette_implementation
from idom.logging import ROOT_LOGGER
from idom.sample import SampleApp as SampleApp
from idom.testing.backend import _hotswap
from idom.testing.display import DisplayFixture


def test_assert_idom_logged_does_not_supress_errors():
    with pytest.raises(RuntimeError, match="expected error"):
        with testing.assert_idom_did_log():
            raise RuntimeError("expected error")


def test_assert_idom_logged_message():
    with testing.assert_idom_did_log(match_message="my message"):
        ROOT_LOGGER.info("my message")

    with testing.assert_idom_did_log(match_message=r".*"):
        ROOT_LOGGER.info("my message")


def test_assert_idom_logged_error():
    with testing.assert_idom_did_log(
        match_message="log message",
        error_type=ValueError,
        match_error="my value error",
    ):
        try:
            raise ValueError("my value error")
        except ValueError:
            ROOT_LOGGER.exception("log message")

    with pytest.raises(
        AssertionError,
        match=r"Could not find a log record matching the given",
    ):
        with testing.assert_idom_did_log(
            match_message="log message",
            error_type=ValueError,
            match_error="my value error",
        ):
            try:
                # change error type
                raise RuntimeError("my value error")
            except RuntimeError:
                ROOT_LOGGER.exception("log message")

    with pytest.raises(
        AssertionError,
        match=r"Could not find a log record matching the given",
    ):
        with testing.assert_idom_did_log(
            match_message="log message",
            error_type=ValueError,
            match_error="my value error",
        ):
            try:
                # change error message
                raise ValueError("something else")
            except ValueError:
                ROOT_LOGGER.exception("log message")

    with pytest.raises(
        AssertionError,
        match=r"Could not find a log record matching the given",
    ):
        with testing.assert_idom_did_log(
            match_message="log message",
            error_type=ValueError,
            match_error="my value error",
        ):
            try:
                # change error message
                raise ValueError("my error message")
            except ValueError:
                ROOT_LOGGER.exception("something else")


def test_assert_idom_logged_assertion_error_message():
    with pytest.raises(
        AssertionError,
        match=r"Could not find a log record matching the given",
    ):
        with testing.assert_idom_did_log(
            # put in all possible params full assertion error message
            match_message=r".*",
            error_type=Exception,
            match_error=r".*",
        ):
            pass


def test_assert_idom_logged_ignores_level():
    original_level = ROOT_LOGGER.level
    ROOT_LOGGER.setLevel(logging.INFO)
    try:
        with testing.assert_idom_did_log(match_message=r".*"):
            # this log would normally be ignored
            ROOT_LOGGER.debug("my message")
    finally:
        ROOT_LOGGER.setLevel(original_level)


def test_assert_idom_did_not_log():
    with testing.assert_idom_did_not_log(match_message="my message"):
        pass

    with testing.assert_idom_did_not_log(match_message=r"something else"):
        ROOT_LOGGER.info("my message")

    with pytest.raises(
        AssertionError,
        match=r"Did find a log record matching the given",
    ):
        with testing.assert_idom_did_not_log(
            # put in all possible params full assertion error message
            match_message=r".*",
            error_type=Exception,
            match_error=r".*",
        ):
            try:
                raise Exception("something")
            except Exception:
                ROOT_LOGGER.exception("something")


async def test_simple_display_fixture():
    if os.name == "nt":
        pytest.skip("Browser tests not supported on Windows")
    async with testing.DisplayFixture() as display:
        await display.show(SampleApp)
        await display.page.wait_for_selector("#sample")


def test_if_app_is_given_implementation_must_be_too():
    with pytest.raises(
        ValueError,
        match=r"If an application instance its corresponding server implementation must be provided too",
    ):
        testing.BackendFixture(app=starlette_implementation.create_development_app())

    testing.BackendFixture(
        app=starlette_implementation.create_development_app(),
        implementation=starlette_implementation,
    )


def test_list_logged_excptions():
    the_error = None
    with testing.capture_idom_logs() as records:
        ROOT_LOGGER.info("A non-error log message")

        try:
            raise ValueError("An error for testing")
        except Exception as error:
            ROOT_LOGGER.exception("Log the error")
            the_error = error

        logged_errors = testing.logs.list_logged_exceptions(records)
        assert logged_errors == [the_error]


async def test_hostwap_update_on_change(display: DisplayFixture):
    """Ensure shared hotswapping works

    This basically means that previously rendered views of a hotswap component get updated
    when a new view is mounted, not just the next time it is re-displayed

    In this test we construct a scenario where clicking a button will cause a pre-existing
    hotswap component to be updated
    """

    def make_next_count_constructor(count):
        """We need to construct a new function so they're different when we set_state"""

        def constructor():
            count.current += 1
            return html.div({"id": f"hotswap-{count.current}"}, count.current)

        return constructor

    @component
    def ButtonSwapsDivs():
        count = Ref(0)

        async def on_click(event):
            mount(make_next_count_constructor(count))

        incr = html.button({"on_click": on_click, "id": "incr-button"}, "incr")

        mount, make_hostswap = _hotswap(update_on_change=True)
        mount(make_next_count_constructor(count))
        hotswap_view = make_hostswap()

        return html.div(incr, hotswap_view)

    await display.show(ButtonSwapsDivs)

    client_incr_button = await display.page.wait_for_selector("#incr-button")

    await display.page.wait_for_selector("#hotswap-1")
    await client_incr_button.click()
    await display.page.wait_for_selector("#hotswap-2")
    await client_incr_button.click()
    await display.page.wait_for_selector("#hotswap-3")



from __future__ import annotations

import ast
import sys
from pathlib import Path

import click

from idom import html
from idom._console.ast_utils import (
    ChangedNode,
    find_element_constructor_usages,
    rewrite_changed_nodes,
)


@click.command()
@click.argument("paths", nargs=-1, type=click.Path(exists=True))
def rewrite_keys(paths: list[str]) -> None:
    """Rewrite files under the given paths using the new html element API.

    The old API required users to pass a dictionary of attributes to html element
    constructor functions. For example:

    >>> html.div({"className": "x"}, "y")
    {"tagName": "div", "attributes": {"className": "x"}, "children": ["y"]}

    The latest API though allows for attributes to be passed as snake_cased keyword
    arguments instead. The above example would be rewritten as:

    >>> html.div("y", class_name="x")
    {"tagName": "div", "attributes": {"class_name": "x"}, "children": ["y"]}

    All snake_case attributes are converted to camelCase by the client where necessary.

    ----- Notes -----

    While this command does it's best to preserve as much of the original code as
    possible, there are inevitably some limitations in doing this. As a result, we
    recommend running your code formatter like Black against your code after executing
    this command.

    Additionally, We are unable to perserve the location of comments that lie within any
    rewritten code. This command will place the comments in the code it plans to rewrite
    just above its changes. As such it requires manual intervention to put those
    comments back in their original location.
    """
    if sys.version_info < (3, 9):  # pragma: no cover
        raise RuntimeError("This command requires Python>=3.9")

    for p in map(Path, paths):
        for f in [p] if p.is_file() else p.rglob("*.py"):
            result = generate_rewrite(file=f, source=f.read_text())
            if result is not None:
                f.write_text(result)


def generate_rewrite(file: Path, source: str) -> str | None:
    tree = ast.parse(source)

    changed = find_nodes_to_change(tree)
    if not changed:
        log_could_not_rewrite(file, tree)
        return None

    new = rewrite_changed_nodes(file, source, tree, changed)
    log_could_not_rewrite(file, ast.parse(new))

    return new


def find_nodes_to_change(tree: ast.AST) -> list[ChangedNode]:
    changed: list[ChangedNode] = []
    for el_info in find_element_constructor_usages(tree, add_props=True):
        for kw in list(el_info.call.keywords):
            if kw.arg == "key":
                break
        else:
            continue

        if isinstance(el_info.props, ast.Dict):
            el_info.props.keys.append(ast.Constant("key"))
            el_info.props.values.append(kw.value)
        else:
            el_info.props.keywords.append(ast.keyword(arg="key", value=kw.value))

        el_info.call.keywords.remove(kw)
        changed.append(ChangedNode(el_info.call, el_info.parents))

    return changed


def log_could_not_rewrite(file: Path, tree: ast.AST) -> None:
    for node in ast.walk(tree):
        if not (isinstance(node, ast.Call) and node.keywords):
            continue

        func = node.func
        if isinstance(func, ast.Attribute):
            name = func.attr
        elif isinstance(func, ast.Name):
            name = func.id
        else:
            continue

        if (
            name == "vdom"
            or hasattr(html, name)
            and any(kw.arg == "key" for kw in node.keywords)
        ):
            click.echo(f"Unable to rewrite usage at {file}:{node.lineno}")



import pytest

from idom import component, config, html
from idom.testing import DisplayFixture, poll
from idom.utils import Ref
from tests.tooling.hooks import use_counter, use_toggle


async def test_script_mount_unmount(display: DisplayFixture):
    toggle_is_mounted = Ref()

    @component
    def Root():
        is_mounted, toggle_is_mounted.current = use_toggle(True)
        return html.div(
            html.div({"id": "mount-state", "data_value": False}),
            HasScript() if is_mounted else html.div(),
        )

    @component
    def HasScript():
        return html.script(
            """() => {
                const mapping = {"false": false, "true": true};
                const mountStateEl = document.getElementById("mount-state");
                mountStateEl.setAttribute(
                    "data-value", !mapping[mountStateEl.getAttribute("data-value")]);
                return () => mountStateEl.setAttribute(
                    "data-value", !mapping[mountStateEl.getAttribute("data-value")]);
            }"""
        )

    await display.show(Root)

    mount_state = await display.page.wait_for_selector("#mount-state", state="attached")
    poll_mount_state = poll(mount_state.get_attribute, "data-value")

    await poll_mount_state.until_equals("true")

    toggle_is_mounted.current()

    await poll_mount_state.until_equals("false")

    toggle_is_mounted.current()

    await poll_mount_state.until_equals("true")


async def test_script_re_run_on_content_change(display: DisplayFixture):
    incr_count = Ref()

    @component
    def HasScript():
        count, incr_count.current = use_counter(1)
        return html.div(
            html.div({"id": "mount-count", "data_value": 0}),
            html.div({"id": "unmount-count", "data_value": 0}),
            html.script(
                f"""() => {{
                    const mountCountEl = document.getElementById("mount-count");
                    const unmountCountEl = document.getElementById("unmount-count");
                    mountCountEl.setAttribute("data-value", {count});
                    return () => unmountCountEl.setAttribute("data-value", {count});;
                }}"""
            ),
        )

    await display.show(HasScript)

    mount_count = await display.page.wait_for_selector("#mount-count", state="attached")
    poll_mount_count = poll(mount_count.get_attribute, "data-value")

    unmount_count = await display.page.wait_for_selector(
        "#unmount-count", state="attached"
    )
    poll_unmount_count = poll(unmount_count.get_attribute, "data-value")

    await poll_mount_count.until_equals("1")
    await poll_unmount_count.until_equals("0")

    incr_count.current()

    await poll_mount_count.until_equals("2")
    await poll_unmount_count.until_equals("1")

    incr_count.current()

    await poll_mount_count.until_equals("3")
    await poll_unmount_count.until_equals("2")


async def test_script_from_src(display: DisplayFixture):
    incr_src_id = Ref()
    file_name_template = "__some_js_script_{src_id}__.js"

    @component
    def HasScript():
        src_id, incr_src_id.current = use_counter(0)
        if src_id == 0:
            # on initial display we haven't added the file yet.
            return html.div()
        else:
            return html.div(
                html.div({"id": "run-count", "data_value": 0}),
                html.script(
                    {
                        "src": f"/_idom/modules/{file_name_template.format(src_id=src_id)}"
                    }
                ),
            )

    await display.show(HasScript)

    for i in range(1, 4):
        script_file = config.IDOM_WEB_MODULES_DIR.current / file_name_template.format(
            src_id=i
        )
        script_file.write_text(
            f"""
            let runCountEl = document.getElementById("run-count");
            runCountEl.setAttribute("data-value", {i});
            """
        )

        incr_src_id.current()

        run_count = await display.page.wait_for_selector("#run-count", state="attached")
        poll_run_count = poll(run_count.get_attribute, "data-value")
        await poll_run_count.until_equals("1")


def test_script_may_only_have_one_child():
    with pytest.raises(ValueError, match="'script' nodes may have, at most, one child"):
        html.script("one child", "two child")


def test_child_of_script_must_be_string():
    with pytest.raises(ValueError, match="The child of a 'script' must be a string"):
        html.script(1)


def test_script_has_no_event_handlers():
    with pytest.raises(ValueError, match="do not support event handlers"):
        html.script({"on_event": lambda: None})


def test_simple_fragment():
    assert html._() == {"tagName": ""}
    assert html._(1, 2, 3) == {"tagName": "", "children": [1, 2, 3]}
    assert html._({"key": "something"}) == {"tagName": "", "key": "something"}
    assert html._({"key": "something"}, 1, 2, 3) == {
        "tagName": "",
        "key": "something",
        "children": [1, 2, 3],
    }


def test_fragment_can_have_no_attributes():
    with pytest.raises(TypeError, match="Fragments cannot have attributes"):
        html._({"some_attribute": 1})
